/* Der er lnt en del kode fra Pthreads1 fra 
https://computing.llnl.gov/tutorials/pthreads/#Misc  */

#include <pthread.h>
#include <stdio.h>

#define NUM_THREADS 5   
#define STAFET_LIMIT 5  

int     count = STAFET_LIMIT;
int     thread_ids[NUM_THREADS];
int		ready_threads = 0;
pthread_mutex_t stafet_mutex;
pthread_mutex_t ready_threads_mutex;
pthread_cond_t stafet_cv[NUM_THREADS];

void *work(void *idp) 
{
  int j,i;
  int *my_id = idp;

  /* bruger vi egentlig mutexen? - udover til cond_wait fordi man SKAL locke fr wait? */
  pthread_mutex_lock(&stafet_mutex);

  while(count >= 0) {
	pthread_mutex_lock(&ready_threads_mutex);
	ready_threads++;
	pthread_mutex_unlock(&ready_threads_mutex);
	
    pthread_cond_wait(&stafet_cv[*my_id], &stafet_mutex);
        
    if(*my_id == 0 && count > 0) { /* Det er den frste trd, og vi skal "tlle stafetten ned" */
      count--;
    }  
    
    if(count == 0) { /* Vi har fet stoppet staffeten, og skal terminere */
      printf("work(): thread %d, stafet count = %d. Terminerer trd.\n", *my_id, count);
      pthread_cond_signal(&stafet_cv[(*my_id+1)%NUM_THREADS]);
      pthread_mutex_unlock(&stafet_mutex);
      pthread_exit(NULL);
    }

    printf("work(): thread %d, stafet count = %d. Signalerer trd %d\n", *my_id, count, ((*my_id+1)%NUM_THREADS)); 
    pthread_cond_signal(&stafet_cv[(*my_id+1)%NUM_THREADS]);
  }
}

int main (int argc, char *argv[])
{
  int i, rc;
  pthread_t threads[NUM_THREADS];
  pthread_attr_t attr;

  /* Initialiser mutex */
  pthread_mutex_init(&stafet_mutex, NULL);
  pthread_mutex_init(&ready_threads_mutex, NULL);

  /* For portabilitet, laver vi alle trde joinable */
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
  
  for (i=0; i<NUM_THREADS; i++) {
	  thread_ids[i] = i;
	  pthread_cond_init (&stafet_cv[i], NULL);
	  pthread_create(&threads[i], &attr, work, (void *)&thread_ids[i]);
  }
  
  /* sov lidt, s vi er sikre p at trd 0 nr ind i while-loopet til sit cond_wait... */
  /*sleep(1);*/
  
  /* Venter p at alle trde er klar */
  while(ready_threads != NUM_THREADS);  /* Her kunne pthread_yield() bruges til at afgive CPU-tid */

  /* signaler trd 0 - start stafetten */
  
  pthread_cond_signal(&stafet_cv[0]);
  
  /* Venter p at alle trdene afslutter */
  for (i=0; i<NUM_THREADS; i++) {
    pthread_join(threads[i], NULL);
  }

  printf ("Main(): Done.\n");

  /* Rydder op og afslutter */
  pthread_attr_destroy(&attr);
  pthread_mutex_destroy(&stafet_mutex);
  pthread_mutex_destroy(&ready_threads_mutex);
  for(i = 0; i < NUM_THREADS; i++)
    pthread_cond_destroy(&stafet_cv[i]);
  pthread_exit(NULL);
}
