#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include "wqueue.h"

/* bruges i wqueue_ts_insert som flag for om ken er tom */
int empty = 0; 
/* bruges i wqueue_ts_run for at holde styr p om der er nogen trde der er igang med arbejde */
int workcounter = 0;
/* mutex til at beskytte ken */
pthread_mutex_t q_mutex;
/* mutex til at beskytte workcounter */
pthread_mutex_t workInProg_mutex;
/* condition til at signalerer at der er arbejde til evt. ventende trde */
pthread_cond_t workInQ_cv;

typedef struct {
    work_f func;
    void *data;
} work_desc_t;

int wqueue_insert(wqueue_t *wq, unsigned int pri, work_f func, void *data) {
    work_desc_t *wd;

    wd = malloc(sizeof(work_desc_t));
    if (wd == NULL) {
        return -1;
    }

    wd->func = func;
    wd->data = data;

    if (pqueue_insert((pqueue_t *)wq, pri, wd) == -1) {
        free(wd);
        return -1;
    }

    return 0;
}


int wqueue_ts_insert(wqueue_t *wq, unsigned int pri, work_f func, void *data) {

  /* ls adgang til ken */
  pthread_mutex_lock(&q_mutex);

  if (((pqueue_t *)wq)->e == NULL) {
    empty = 1;
  }

  work_desc_t *wd;

    wd = malloc(sizeof(work_desc_t));
    if (wd == NULL) {
      return -1;
    }

    wd->func = func;
    wd->data = data;
    
    if (pqueue_insert((pqueue_t *)wq, pri, wd) == -1) {
        free(wd);
        return -1;
    }
    printf("wqueue_ts_insert(): Der er indsat arbejde i ken\n");
    /* check for om ken var tom - jvf. opgaveteksten */
    if(empty) {
      printf("wqueue_ts_insert(): Ken var tom, men har nu noget i igen - Der broadcastes\n");
      /* pthread_mutex_lock(&q_mutex); */
      pthread_cond_broadcast(&workInQ_cv);
      /* pthread_mutex_unlock(&q_mutex); */
      empty = 0;
    }

    /* ls op for ken */
    pthread_mutex_unlock(&q_mutex);
    return 0;
}


/* Det er denne funktion trdene fra threadpool krer - den udtager arbejde hvis der er noget,
   venter hvis ken er tom, men andre trde stadig arbejder og
   afslutter hvis der hverken er arbejdene trde eller noget i ken
*/
void *wqueue_ts_run(void *wq) {

  /* bool til at holde styr p om ken er tom */
  int not_empty;

  pthread_mutex_lock(&q_mutex);
  /* vi tjekker at der er noget i ken */
  not_empty = (((pqueue_t *)wq)->e != NULL);

  while(not_empty || workcounter > 0) {   /* s lnge der evt. er noget at lave... */
     
    printf("wqueue_ts_run(): workcounter: %d. Trd_id: %d. K !tom?: %d \n", workcounter, pthread_self(), not_empty);
      
    if(not_empty) {
      /* begynd arbejde */
      pthread_mutex_lock(&workInProg_mutex);
      workcounter++;
      pthread_mutex_unlock(&workInProg_mutex);
      
      /* Tag arbejde ud af ken */
      work_desc_t *wd;
      wd = pqueue_remove((pqueue_t *)wq);
      
      /* unlock ken da vi ikke arbejder p den og udfr arbejdet */
      pthread_mutex_unlock(&q_mutex);
      wd->func(wd->data);
      free(wd);
     
      /* arbejde slut */
      pthread_mutex_lock(&workInProg_mutex);
      printf("wqueue_ts_run(): Trd %d har udtaget og udfrt arbejde\n", pthread_self());
      workcounter--;
      pthread_mutex_unlock(&workInProg_mutex);

      /* vi locker ken igen - de trde der gr ind i denne if unlockede den
	 tidligere. Det nste der sker er nemlig et check p ken*/
      pthread_mutex_lock(&q_mutex);
    }
    
    /*tjek om der er noget i ken */
    not_empty = (((pqueue_t *)wq)->e != NULL);

    /* Vent p mere arbejde - hvis ken er tom og der stadig er nogen der arbejder*/
    if((!not_empty) && (workcounter > 0)) {
      printf("wqueue_ts_run(): Trd %d venter p arbejde\n",pthread_self());
      pthread_cond_wait(&workInQ_cv, &q_mutex);
      printf("wqueue_ts_run(): Trd %d er signaleret\n", pthread_self());
    }
    
    /*tjek om der er noget i ken */
    not_empty = (((pqueue_t *)wq)->e != NULL); 
  }

  /* Der er ikke mere arbejde */ 
  printf("wqueue_ts_run(): Trd_id: %d signing off\n", pthread_self());
  
  /* signaler evt. andre ventende trde */
  pthread_cond_broadcast(&workInQ_cv);
  pthread_mutex_unlock(&q_mutex);
  
  pthread_exit(NULL);
}


int wqueue_thread_pool(wqueue_t *wq, unsigned int cnt) {

  int i;
  pthread_attr_t attr;
  pthread_t pthreads[cnt];

  /* Initialiser mutex og condition variabel */
  pthread_mutex_init(&q_mutex, NULL);
  pthread_mutex_init(&workInProg_mutex, NULL);
  pthread_cond_init(&workInQ_cv, NULL);

  /* For portabilitet, laver vi alle trde joinable */
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

  /* initialiser trde */
  for (i=0; i<cnt; i++) {
    pthread_create(&pthreads[i], &attr, wqueue_ts_run, (void *) wq);
  }
  printf("wqueue_thread_pool(): Trde oprettet - arbejd!...\n");

  /* Venter p at alle trdene afslutter */
  for (i=0; i<cnt; i++) {
    pthread_join(pthreads[i], NULL);
  }

  /* Rydder op og afslutter */
  pthread_attr_destroy(&attr);
  pthread_cond_destroy(&workInQ_cv);

  return 0;
}


// initialiser en k
pqueue_t *init_queue()
{
  pqueue_t *temp;
  temp = malloc(sizeof(pqueue_t));
  if (temp == NULL)
    exit(-1);
  temp->e = NULL;
  return temp;
}

